"""
Address Information.

Access can be done to addrspacesa, words and fields,
and addresses in a single, burst or scatter flavour using offsets, masks, etc.
:any:`create` takes all these variants into account and serves a normalized information set `AddrInfo`.
"""

from collections.abc import Iterator
from typing import TypeAlias

import ucdp as u
from pydantic import PositiveInt

from .addrmap import AddrMap
from .addrmapref import AddrMapRef, ToAddrMapRef
from .addrrange import AddrRange
from .resolver import resolve

Size: TypeAlias = PositiveInt
Depth: TypeAlias = PositiveInt


class AddrInfo(u.Object):
    """Address Info."""

    ref: AddrMapRef
    addrrange: AddrRange
    mask: u.Hex | None = None

    def __str__(self) -> str:
        info = f"{self.ref!s} [{self.addrrange!s}]"
        if self.mask is not None:
            info = f"{info} mask={self.mask}"
        return info

    @staticmethod
    def create(addrmap: AddrMap, item: ToAddrMapRef, offset: int | None = None, mask: int | None = None) -> "AddrInfo":
        """
        Create `AddrInfo`.

        Args:
            addrmap: Address Map
            item: Thing to be resolved

        Keyword Args:
            offset: address offset in bytes
            mask: Value Mask (not allowed for fields)

        Example:
            >>> import ucdp as u
            >>> from ucdp_addr import Addrspace, Word, Field, AddrMap
            >>> addrmap = AddrMap()
            >>> for idx, aname in enumerate(("uart", "spi", "owi")):
            ...     addrspace = Addrspace(name=aname, width=32, depth=8, baseaddr=8*4*idx)
            ...     addrmap.add(addrspace)
            ...     word = addrspace.add_word("ctrl")
            ...     field = word.add_field("ena", u.BitType(), "RW")
            ...     word = addrspace.add_word("stat")
            ...     field = word.add_field("bsy", u.BitType(), "RO", offset=9)

        Addrspace:

            >>> addrinfo = AddrInfo.create(addrmap, "spi")
            >>> str(addrinfo)
            'spi [0x20 8x32 (32 bytes)]'
            >>> addrinfo.ref
            AddrMapRef(..., addrspace=Addrspace(name='spi', baseaddr=Hex('0x20'), size=Bytesize('32 bytes')))
            >>> addrinfo.mask
            >>> tuple(addrinfo.iter())
            (Hex('0x20'), Hex('0x24'), Hex('0x28'), Hex('0x2C'), Hex('0x30'), Hex('0x34'), Hex('0x38'), Hex('0x3C'))

            >>> addrinfo = AddrInfo.create(addrmap, "spi", offset=8, mask=0xF0)
            >>> str(addrinfo)
            'spi [0x28 6x32 (24 bytes)] mask=0xF0'
            >>> addrinfo.mask
            Hex('0xF0')
            >>> tuple(addrinfo.iter())
            (Hex('0x28'), Hex('0x2C'), Hex('0x30'), Hex('0x34'), Hex('0x38'), Hex('0x3C'))

        Word:

            >>> addrinfo = AddrInfo.create(addrmap, "spi.stat")
            >>> str(addrinfo)
            'spi.stat [0x24 1x32 (4 bytes)]'
            >>> addrinfo.ref
            AddrMapRef(..., addrspace=Addrspace(name='spi', baseaddr=Hex('0x20'), ..., word=Word(name='stat', ...)
            >>> addrinfo.mask
            >>> tuple(addrinfo.iter())
            (Hex('0x24'),)

            >>> AddrInfo.create(addrmap, "spi.stat", offset=8)
            Traceback (most recent call last):
              ...
            ValueError: 'offset' is not allowed for words and fields

        Field:

            >>> addrinfo = AddrInfo.create(addrmap, "spi.stat.bsy")
            >>> str(addrinfo)
            'spi.stat.bsy [0x24 1x32 (4 bytes)] mask=0x200'
            >>> addrinfo.ref
            AddrMapRef(...Hex('0x20'), ... word=Word(name='stat', ...(name='bsy', type_=BitType(), bus=RO, offset=9))
            >>> addrinfo.mask
            Hex('0x200')
            >>> tuple(addrinfo.iter())
            (Hex('0x24'),)

            >>> AddrInfo.create(addrmap, "spi.stat.bsy", offset=8)
            Traceback (most recent call last):
              ...
            ValueError: 'offset' is not allowed for words and fields
            >>> AddrInfo.create(addrmap, "spi.stat.bsy", mask=8)
            Traceback (most recent call last):
              ...
            ValueError: 'mask' is not allowed for fields
        """
        ref = resolve(addrmap, item)
        addrrange = ref.addrrange
        if ref.field:
            if mask is not None:
                raise ValueError("'mask' is not allowed for fields")
            mask = ref.field.slice.mask

        # apply offset
        if offset is not None:
            addrrange = _apply_offset(ref, addrrange, offset)

        return AddrInfo(ref=ref, addrrange=addrrange, mask=mask)

    def iter(self) -> Iterator[u.Hex]:
        """Iteratate over address ranges."""
        addrrange = self.addrrange
        wordsize = addrrange.wordsize
        for idx in range(addrrange.depth):
            yield addrrange.baseaddr + int(idx * wordsize)


def _apply_offset(ref: AddrMapRef, addrrange: AddrRange, offset: int) -> AddrRange:
    if ref.word or ref.field:
        raise ValueError("'offset' is not allowed for words and fields")
    if offset >= addrrange.size:
        raise ValueError(f"offset={offset!r} exceeds size ({addrrange.size})")
    return addrrange.new(baseaddr=addrrange.baseaddr + offset, size=addrrange.size - offset, depth=None)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Spaces.
"""

from collections.abc import Iterable, Iterator
from logging import getLogger

from .addrspace import Addrspace

LOGGER = getLogger(__name__)

Addrspaces = Iterable[Addrspace]


def join_addrspaces(base: Addrspace, addrspaces: Addrspaces) -> Iterator[Addrspace]:
    """Join Address Spaces."""
    for addrspace in addrspaces:
        joined = base.join(addrspace)
        LOGGER.debug("join_addrspaces: %s+%s=%s", base, addrspace, joined)
        if joined:
            yield joined


def zip_addrspaces(lefts, rights) -> Iterator[tuple[Addrspace, ...]]:
    """
    Zip Address Spaces.

    >>> one = (
    ...     Addrspace(name='a0', baseaddr=0x0000, size=0x1000),
    ...     Addrspace(name='a1', baseaddr=0x1000, size=0x1000))
    >>> other = (
    ...     Addrspace(name='b0', baseaddr=0x0000, size=0x1000),
    ...     Addrspace(name='b1', baseaddr=0x1000, size=0x800),
    ...     Addrspace(name='b2', baseaddr=0x1800, size=0x800))
    >>> for left, right in zip_addrspaces(one, other): print(f"{str(left)!r} {str(right)!r}")
    'a0 +0x0 1024x32' 'b0 +0x0 1024x32'
    'a1 +0x1000 1024x32' 'b1 +0x1000 512x32'
    'a1 +0x1000 1024x32' 'b2 +0x1800 512x32'
    """
    leftiter = iter(lefts)
    rightiter = iter(rights)
    try:
        left = next(leftiter)
        right = next(rightiter)
    except StopIteration:
        pass
    else:
        while True:
            yield left, right
            leftend = left.endaddr <= right.endaddr
            rightend = right.endaddr <= left.endaddr
            if leftend:
                try:
                    left = next(leftiter)
                except StopIteration:
                    break
            if rightend:
                try:
                    right = next(rightiter)
                except StopIteration:
                    break
            assert leftend or rightend



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Space."""

import ucdp as u

from ucdp_addr import AddrMap, Addrspace, Addrspaces, join_addrspaces


class TopMod(u.AMod):
    """Example Top Module."""

    def _build(self):
        SubMod(self, "u_sub0")
        SubMod(self, "u_sub1")

    def get_addrspaces(self, **kwargs) -> Addrspaces:
        """Determine Address Spaces."""
        base = Addrspace(name="main0", baseaddr=0x10000, size=0x10000, is_sub=False)
        yield from join_addrspaces(base, self.get_inst("u_sub0/u_bus").get_addrspaces(**kwargs))

        base = Addrspace(name="main1", baseaddr=0x40000, size=0x10000, is_sub=False)
        yield from join_addrspaces(base, self.get_inst("u_sub1/u_bus").get_addrspaces(**kwargs))


class SubMod(u.AMod):
    """Example Sub Module."""

    def _build(self):
        BusMod(self, "u_bus")
        BusMod(self, "u_otherbus")


class BusMod(u.AMod):
    """Bus Module."""

    addrmap: AddrMap = u.Field(default_factory=AddrMap)

    def _build(self):
        addrspace = Addrspace(name="one", baseaddr=0x1000, size="2KB")
        self.addrmap.add(addrspace)
        word = addrspace.add_word("word0", offset=4)
        word.add_field("field0", u.UintType(4), "RW")

        addrspace = Addrspace(name="two", baseaddr=0x2000, size="4KB")
        self.addrmap.add(addrspace)
        word = addrspace.add_word("word1", offset=0x3F0)
        word.add_field("field1", u.UintType(32), "RO")

    def get_addrspaces(self, master=None, **kwargs) -> Addrspaces:
        """Address Spaces."""
        for addrspace in self.addrmap:
            if master and addrspace.name == "two":
                continue
            yield addrspace


class EmptyMod(u.AMod):
    """No get_addrspaces implementation."""

    def _build(self):
        pass



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Unified Chip Design Platform - Address Map."""

from .access import (
    ACCESSES,
    NA,
    RC,
    RO,
    RP,
    RS,
    RT,
    RW,
    RW0C,
    RW0S,
    RW1C,
    RW1S,
    RWL,
    W0C,
    W0S,
    W1C,
    W1S,
    WL,
    WO,
    Access,
    ReadOp,
    ToAccess,
    WriteOp,
    any_read,
    cast_access,
    get_counteraccess,
    is_read_repeatable,
    is_write_repeatable,
)
from .addrdecoder import AddrDecoder
from .addrinfo import AddrInfo, Depth, Size
from .addrmap import AddrMap, AddrspaceFilter, FillAddrspaceFactory, create_fill_addrspace
from .addrmapfinder import GetAttrspacesFunc, get_addrmap, get_addrspaces
from .addrmapref import AddrMapRef, ToAddrMapRef
from .addrmaster import AddrMaster
from .addrmatrix import AddrMatrix
from .addrrange import AddrRange
from .addrref import AddrRef
from .addrslave import AddrSlave, SlaveAddrspace
from .addrspace import (
    Addrspace,
    DefaultAddrspace,
    Field,
    FieldFilter,
    FillFieldFactory,
    FillWordFactory,
    FullError,
    NamingScheme,
    ReservedAddrspace,
    Word,
    WordFields,
    WordFilter,
    Words,
    create_fill_field,
    create_fill_word,
    get_is_const,
    get_is_volatile,
    get_mask,
    name_alpha,
    read_on_modify,
    resolve_field_value,
)
from .addrspacealias import AddrspaceAlias
from .addrspaces import Addrspaces, join_addrspaces, zip_addrspaces
from .const import NOREF
from .data import DataType, check_data, get_datatype, get_size
from .datainfo import AddrData, DataInfo
from .defines import Defines
from .resolver import resolve, resolves
from .util import calc_depth_size, split_keyvaluepairs
from .wordinfo import WordInfo

__all__ = [
    "Access",
    "ACCESSES",
    "AddrData",
    "AddrDecoder",
    "AddrInfo",
    "AddrMap",
    "AddrmapRef",
    "AddrMapRef",
    "AddrMaster",
    "AddrMatrix",
    "AddrRange",
    "AddrRef",
    "AddrSlave",
    "Addrspace",
    "AddrspaceAlias",
    "AddrspaceFilter",
    "Addrspaces",
    "any_read",
    "calc_depth_size",
    "cast_access",
    "check_data",
    "create_fill_addrspace",
    "create_fill_field",
    "create_fill_word",
    "DataInfo",
    "DataType",
    "DefaultAddrspace",
    "Defines",
    "Depth",
    "Field",
    "FieldFilter",
    "FillAddrspaceFactory",
    "FillFieldFactory",
    "FillWordFactory",
    "FullError",
    "get_addrmap",
    "get_addrspaces",
    "get_counteraccess",
    "get_datatype",
    "get_is_const",
    "get_is_volatile",
    "get_mask",
    "get_size",
    "GetAttrspacesFunc",
    "is_read_repeatable",
    "is_write_repeatable",
    "join_addrspaces",
    "NA",
    "name_alpha",
    "NamingScheme",
    "NOREF",
    "RC",
    "read_on_modify",
    "ReadOp",
    "ReservedAddrspace",
    "resolve_field_value",
    "resolve",
    "resolves",
    "RO",
    "RP",
    "RS",
    "RT",
    "RW",
    "RW0C",
    "RW0S",
    "RW1C",
    "RW1S",
    "RWL",
    "Size",
    "SlaveAddrspace",
    "split_keyvaluepairs",
    "ToAccess",
    "ToAddrMapRef",
    "W0C",
    "W0S",
    "W1C",
    "W1S",
    "WL",
    "WO",
    "Word",
    "WordFields",
    "WordFilter",
    "WordInfo",
    "Words",
    "WriteOp",
    "zip_addrspaces",
]



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Range.
"""

import ucdp as u
from icdutil import num
from pydantic import PositiveInt

from .util import calc_depth_size

BASEADDR_DEFAULT = u.Hex(0)


class AddrRange(u.Object):
    """
    Address Range.

    Examples:
        >>> addrrange = AddrRange(size='4k')
        >>> addrrange
        AddrRange(size=Bytesize('4 KB'))
        >>> addrrange.baseaddr
        Hex('0x0')
        >>> addrrange.width
        32
        >>> addrrange.depth
        1024
        >>> addrrange.size
        Bytesize('4 KB')
        >>> addrrange.addrwidth
        12
        >>> addrrange.endaddr
        Hex('0xFFF')
        >>> addrrange.nextaddr
        Hex('0x1000')
        >>> addrrange.wordsize
        4.0
        >>> str(addrrange)
        '0x0 1024x32 (4 KB)'
        >>> str(AddrRange(depth=16))
        '0x0 16x32 (64 bytes)'
        >>> str(AddrRange(size=16))
        '0x0 4x32 (16 bytes)'
    """

    baseaddr: u.Hex = BASEADDR_DEFAULT
    """Base Address"""
    width: PositiveInt = 32
    """Width in Bits."""
    depth: PositiveInt = u.Field(repr=False)
    """Number of words."""
    size: u.Bytes
    """Size in Bytes."""

    def __init__(
        self,
        baseaddr: u.Hex = BASEADDR_DEFAULT,
        width: PositiveInt = 32,
        depth: PositiveInt | None = None,
        size: u.Bytes | None = None,
        **kwargs,
    ):
        depth, size = calc_depth_size(width, depth, size)
        super().__init__(baseaddr=baseaddr, width=width, depth=depth, size=size, **kwargs)

    @property
    def addrwidth(self) -> PositiveInt:
        """Address Width."""
        return num.calc_unsigned_width(int(self.size) - 1)

    @property
    def addrmask(self) -> PositiveInt:
        """Address Width."""
        return (2**self.addrwidth) - 1

    @property
    def endaddr(self) -> u.Hex:
        """End Address - `baseaddr+size-1`."""
        return self.baseaddr + self.size - 1

    @property
    def nextaddr(self) -> u.Hex:
        """Next Free Address - `baseaddr+size`."""
        return self.baseaddr + self.size

    @property
    def wordsize(self) -> float:
        """Number of Bytes Per Word."""
        return self.width / 8

    def __str__(self) -> str:
        return f"{self.baseaddr} {self.org}"

    @property
    def org(self) -> str:
        """Organization."""
        return f"{self.depth}x{self.width} ({self.size})"
